"""
Creation:
    Author: Martin Grunnill
    Date: 2024-03-08
Description: Generate Latin-Hypercube Sample (LHS), run simulations and calculate Partial
Correlation Coefficients (PRCCs). For a description of LHS and PRCCs use in model sensitivity analyses see:
    Marino, S., Hogue, I. B., Ray, C. J., & Kirschner, D. E. (2008). A methodology for performing global uncertainty and
    sensitivity analysis in systems biology. In Journal of Theoretical Biology (Vol. 254, Issue 1, pp. 178–196).
    https://doi.org/10.1016/j.jtbi.2008.04.011

Notes
-----
Serial processing of LHS can be slow. but parallel processing of LHS can take up a lot of computing resources.

"""
import pandas as pd
import pingouin as pg
from dask.distributed import Client
from scipy.stats import qmc
from tqdm.auto import tqdm


def _format_sample(parameters_df, LH_samples, other_samples_to_repeat=None):
    """
    Formats Latin Hypercube sample generated by scipy.stats.qmc.

    Scales LH_samples in with boundaries outlined in parameters_df.

    Parameters
    ----------
    parameters_df : pd.DataFrame
        DataFrame outlining the boundaries for each parameter. Must contain fields 'Lower Bound' and
        'Upper Bound'. Name of the parameters is assumed to be in the index.
    LH_samples : numpy.array
        Output from scipy.stats.qmc.LatinHypercube.
    other_samples_to_repeat : pandas.DataFrame
        Samples to resampled and merged with LH samples.

    Returns
    -------
    samples_df : pandas.Dataframe
        Fully formatted samples.
    parameters_sampled : list
        A list of samples being sampled.
    """
    # if any(~parameters_df['Distribution'].isin(['boolean','uniform'])):
    #     raise ValueError('Only Boolean and Uniform distributions currently supported.')
    samples_df = pd.DataFrame(qmc.scale(LH_samples,
                                        parameters_df['Lower Bound'],
                                        parameters_df['Upper Bound']),
                              columns=parameters_df.index)
    if other_samples_to_repeat is not None:
        multiple = len(samples_df) / len(other_samples_to_repeat)
        if not multiple.is_integer():
            raise ValueError(
                'LHS sample size devided by length of other_samples_to_repeat must be expressable as an interger value.')
        multiple = int(multiple)
        repeated_samples = pd.concat([other_samples_to_repeat] * multiple, ignore_index=True)
        samples_df = pd.concat([samples_df, repeated_samples], axis=1, ignore_index=True)
        parameters_sampled = parameters_df.index.to_list() + other_samples_to_repeat.columns.to_list()
        samples_df.columns = parameters_sampled
    else:
        parameters_sampled = samples_df.columns.to_list()

    # convert_to_bool = parameters_df.Parameter[parameters_df['Distribution'] == 'boolean']
    # for parameter in convert_to_bool:
    #     samples_df[parameter] = samples_df[parameter] >= 0.5
    return samples_df, parameters_sampled


def calculate_prcc(results_and_sample_df, parameter, output, covariables, method='spearman'):
    """
    Calculate Partial Correlation Coefficient PCC (default Rank 'Spearman' PRCC).

    A wrapper for pingouin.partial_corr. Partial Rank Correlation Coefficients (PRCCS) can be used to evaluate
    sensitivity of a model to a parameter[1].

    Parameters
    ----------
    results_and_sample_df : pandas.DataFrame
        DataFrame of results and parameters for calculating PCC.
    parameter : string
        Parameter for which PCC will be calculated.
    output : string
        Model output for which PCC will be calculated.
    covariables : list of strings
        Parameters whose effects will be discounted.
    method : string, default 'spearman' (Rank correlations)
        Form of PCC see documentation of pingouin.partial_corr.


    Returns
    -------
    pandas.DataFrame
        Partial Corelation Coefficient of parameter and output.

    References
    ----------
    [1] Marino, S., Hogue, I. B., Ray, C. J., & Kirschner, D. E. (2008). A methodology for performing global uncertainty
        and sensitivity analysis in systems biology. In Journal of Theoretical Biology (Vol. 254, Issue 1, pp. 178–196).
        https://doi.org/10.1016/j.jtbi.2008.04.011
    """
    param_pcor = pg.partial_corr(results_and_sample_df,
                                 x=parameter,
                                 y=output,
                                 covar=covariables,
                                 method=method)
    param_pcor.rename(index={method: parameter + ' on ' + output}, inplace=True)
    param_pcor['Number_of_Covariables'] = len(covariables)
    confidence_interval = pd.DataFrame(param_pcor['CI95%'].tolist())
    param_pcor['lower_CI_0.95'] = confidence_interval[0][0]
    param_pcor['upper_CI_0.95'] = confidence_interval[1][0]
    param_pcor.drop(columns=['CI95%'], inplace=True)
    return param_pcor


def lhs_prcc(parameters_df,
             sample_size,
             model_run_method,
             client=None,
             lhs_obj=None,
             other_samples_to_repeat=None,
             **kwargs):
    """
    Generate a Latin Hypercube sample, run model with sample and calculate PRCC for sampled parameters.

    Latin Hypercube Sampling with Partial Rank Correlation Coefficients (PRCCS) can be used to evaluate sensitivity of a
    model to a parameter[1]. Note currently only supports uniform distribution.

    Parameters
    ----------
    parameters_df : pandas.DataFrame
        DataFrame outlining the boundaries for each parameter. Must contain fields 'Lower Bound' and
        'Upper Bound'. The name of the parameters is assumed to be in the index of the DataFrame.
    sample_size : int
        Sample size of Latin Hypercube.
    model_run_method : function
        Method of running model's simulations. Must accept parameters as a single dictionary. Must output dictionary of 
        input parameters and model results.      
    client: dask.distributed.Client (default None)
        Dask client for running simulations in parallel. If not given simulations are run serially with a tqdm progress
         bar.
    lhs_obj : scipy.stats.qmc.LatinHypercube, optional
        Pre-initialised Latin Hypercube sample generator. If not provided one is generated by within this function.
    other_samples_to_repeat : pandas.DataFrame
        Samples to resampled and merged with LH samples.
    kwargs: 
        Key word arguments to be passed to model_run_method.

    Returns
    -------
    results_df : pandas.DataFrame
        Results of simulations preceded by parameters used in simulations.
    prccs : pandas.DataFrame
        PRCC summary of the model's sensitivity to its parameters.

    References
    ----------
    [1] Marino, S., Hogue, I. B., Ray, C. J., & Kirschner, D. E. (2008). A methodology for performing global uncertainty
        and sensitivity analysis in systems biology. In Journal of Theoretical Biology (Vol. 254, Issue 1, pp. 178–196).
        https://doi.org/10.1016/j.jtbi.2008.04.011
    """
    if not pd.api.types.is_object_dtype(parameters_df.index):
        raise TypeError('Index of parameters_df should be the models parameters. Therefore, they should be strings.')
    if lhs_obj is None:
        num_LH_parameters_sampled = len(parameters_df)
        lhs_obj = qmc.LatinHypercube(num_LH_parameters_sampled)
    LH_sample = lhs_obj.random(sample_size)
    sample_df, parameters_sampled = _format_sample(parameters_df, LH_sample, other_samples_to_repeat)
    if client is None:
        results = run_samples_serially(sample_df, model_run_method, **kwargs)
    else:
        results = run_samples_in_parallel(sample_df, model_run_method, client=client, **kwargs)

    results_df = pd.DataFrame.from_records(results)
    result_columns = [column
                      for column in list(results_df.columns)
                      if column not in parameters_sampled]
    results_df = results_df[parameters_sampled + result_columns]

    prccs = []
    ### FUTURE IMPROVEMENT ###
    # Is it worth running nested for loops below in parallel if given dask client? It already runs very quickly.
    # However, I have never had to many calls of calculate_prcc to run.
    for parameter in parameters_sampled:
        covariables = [item
                       for item in parameters_sampled
                       if item != parameter]
        for column in result_columns:
            param_rank_pcor = calculate_prcc(results_and_sample_df=results_df,
                                             parameter=parameter,
                                             output=column,
                                             covariables=covariables,
                                             method='spearman')
            prccs.append(param_rank_pcor)

    prccs = pd.concat(prccs)
    prccs.sort_index(inplace=True)
    return results_df, prccs


def run_samples_serially(parameters_df, model_run_method, **kwargs):
    results = []
    for parameters in tqdm(parameters_df.to_dict('records'),
                           desc='Simulating LH Sample',
                           position=1,
                           leave=False,
                           colour='green'):
        results.append(model_run_method(parameters, **kwargs))
    return results


def run_samples_in_parallel(parameters_df, model_run_method, client, **kwargs):
    """
    Runs model simulations using parameters in sample_df using parallel processing.

    Parameters
    ----------
    parameters_df : pandas.Dataframe
        Parameter samples being run. Column fields are parameters.
    model_run_method : function
        Method of running model's simulations. Must accept parameters as a single dictionary in the first argument. Must
         output dictionary of input parameters and model results.
    kwargs : 
        Key word arguments to pass to model_run_method.

    Returns
    -------
    If return_focused_results is True a pandas DataFrame of results is returned.
    """
    if not isinstance(client, Client):
        raise TypeError('client must be a dask.distributed.Client.')
    ### FUTURE IMPROVEMENT ###
    # Add progress bar  below so users do not have to use the one on the dask client dashboard.
    futures = []
    for parameters in parameters_df.to_dict('records'):
        future = client.submit(model_run_method, parameters, **kwargs)
        futures.append(future)

    results = client.gather(futures)
    return results

if __name__ == "__main__":
    pass




"""
Creation:
    Author: Martin Grunnill
    Date: 2024/03/08
Description: 
    Contains class for setting up and simulating multidimensional metapopulation models.

Classes
-------
MetaCaster
    Class for setting up and simulating multidimensional metapopulation models.

    Can simulate models on its own if given model_attributes and subpop_model arguments at initialisation.
    Alternatively these can be defined as attributes of MetaCaster subclass.

Notes
-----
I have tried to use notation find in:
    Keeling, M. J., & Rohani, P. (2008). Metapopulations. In Modeling Infectious Diseases in Humans and Animals
    (pp. 237–240). Princeton University Press.

"""
import copy
from inspect import getfullargspec
from collections.abc import Iterable
import itertools
import numpy as np
import pandas as pd
from numbers import Number
import scipy


def _check_string_in_list_strings(string, list_strings):
    if not isinstance(string, str):
        raise TypeError(str(string) + ' should be of type string.')

    check_list = eval('self.' + list_strings)
    if string not in check_list:
        raise ValueError(string + ' is not one of the predefined model ' + list_strings + ': ' +
                         ','.join(check_list[:-1]) + ' and ' + check_list[:-1] + '.')

def _nested_dict_values(d):
    return [index for sub_d in d.values() for index in sub_d.values()]


def _unionise_dict_of_lists(dict_of_lists):
    """
    Merge dictionary value lists into a list of unique values.

    Parameters
    ----------
    dict_of_lists : dictionay of lists

    Returns
    -------
    list
        Merged unique list values.
    """
    return list(set().union(*dict_of_lists.values()))


def _is_iterable_of_unique_elements(object):
    if not isinstance(object, Iterable):
        return False
    elif len(object) > len(set(object)):
        return False
    else:
        return True


def _collection_of_strings(object):
    if isinstance(object, (dict, str)):
        return False
    elif any(not isinstance(element, str) for element in object):
        return False
    else:
        return True


def _is_set_like_of_strings(object):
    if not _is_iterable_of_unique_elements(object):
        return False
    else:
        return _collection_of_strings(object)


def select_dict_items_in_list(dictionary, lst):
    return {key: value for key, value in dictionary.items() if key in lst}


class MetaCaster:
    """
    Class for setting up and simulating multidimensional metapopulation models.

    Can simulate models on its own if given model_attributes and subpop_model arguments at initialisation.
    Alternatively these can be defined as attributes of MetaCaster subclass.

    Parameters
    ----------
    dimensions : int, collection of unique strings, list/tuple of ints, list/tuple of collections of unique strings OR list/tuple of dictionaries (transfer dictionaries)
        If int :
            This creates a one dimension metapopulation structure with range(scaffold) used to label subpopulations
            in dimensions attribute.
        If list/tuple/set of unique strings:
            This creates a one dimension metapopulation structure with entries used to label subpopulations
             in dimensions attribute.
        If list/tuple of ints:
             This creates a multidimensional metapopulation structure with range(each int entry) used to generate
              labels on an axis of the subpopulations in dimensions attribute.
        If list/tuple of list/tuple/set of unique strings:
         This creates a multidimensional metapopulation structure with each sub-list/tuple/set entries used as
          labels in am axis of dimensions attribute.
        If  list/tuple of dictionaries (transfer dictionaries):
            Transfer dictionaries outlines the transfer of one subpopulation to another subpopulation.
            Each transfer dictionary must have the key values pairs:
                from_coordinates: string/int or list/tuple of strings/ints
                    Subpopulation coordinates from which hosts are leaving. All of these entries should be of the same
                     length.
                to_coordinates: string/int or list/tuple of strings/ints
                    Subpopulation coordinates from which hosts are leaving.All of these entries should be of the same
                     length and the same length as the from_coordinates entries.
                states: list of strings or string
                    Host states which will transition between subpopulations. Single entry of 'all' value
                    means all the available model states transition between subpopulations. Alternatively a list of
                    specific states can be given.
                parameter : string
                    Name given to parameter that is responsible for flow of hosts transferring between subpopulations.
            Optional key value pairs:
                piecewise targets: list, tuple, numpy.Array or pandas.Series
                    Targets for piecewise estimation of parameter that is responsible for flow of hosts transitions
                    between clusters and vaccine groups (see method group_transfer).
    subpop_model : callable function/class method
        Method used to model subpopulation:
            Required arguments: 'y', 'y_deltas', 'parameters' and 'states_index'.
            Possible additional arguments: 'coordinates', 'subpop_suffix', 'foi' or 't'.
    other_model_attributes : kwargs
        Named arguments are attributes to be set. Alternatively these attributes can be set in a subclass of MetaCaster.

    Attributes
    ----------
    states : list of strings
        States used in model.
    observed_states : list of strings
        Observed states. Useful for obtaining results or fitting (e.g. Cumulative incidence).
    infected_states : list of strings
        Infected states (not necessarily infectious).
    infectious_states : list of strings
        Infectious states. These states contribute to force of infection. If None no force of infection is
        caluculated when running model (within ode method).
    symptomatic_states : list of strings
        Symptomatic states. NOTE any state in the list self.infectious_states but NOT in this list has its transmission
        modified by self.asymptomatic_transmission_modifier (see method calculate_fois).
    transmission_term_prefix : string, default 'beta'
        Prefix of subpopulation term used in calculating forces of infection, default term is 'beta' (see method
        calculate_fois).
    population_term_prefix : string, default 'N'
        Prefix of subpopulation term used as denominator in calculating force of infection (see method calculate_fois).
    subpop_interaction_prefix : string default is 'rho'
        Prefix of subpopulation term used to denote level of interaction between subpopulations when calculating force
         of infection (see method calculate_fois).
    asymptomatic_transmission_modifier : string or None
        Factor used to modify transmission from infectious but asymptomatic states. If None a factor of 1 is used in the
        asymptomatic_transmission_modifier's place when using method calculating_fois.
    universal_params : list of strings
        A list of all the parameters that are NOT:
        - directly to do with transmission
        - subpopulation specific (sub_pop_params).
    subpop_params : list of strings, or if axis/dimensions > 1 a dictionary {int: lists of strings}
         A list of all the parameters that axis/dimensions specific but NOT directly to do with transmission.
    parameter_names : list of strings
        The names of all the parameters.
    population_terms : list of strings
        Subpopulation population terms used in calculating force of infection (see method calculate_fois).
    transmission_terms :  list of strings
        The transmission terms for each subpopulation (see method calculate_fois).
    subpop_interaction_terms : list of strings
        The terms for each interaction between subpopulations used when calculating force of infection
        (see method calculate_fois).
    total_subpops : int
        Total number of subpopulations.
    total_parameters : int
        Total number of parameters in model.
    dimensions : list of sets of ints/strings
        Dimensions of metapopulation.
    subpop_transfer_dict : dictionary {tuple of ints or strings: list of dictionaries}
        Each entry outlines all the outflows from a subpopulation to other subpopulation. For use with
        subpop_transfer method.
    subpop_transition_params_dict : dictionary {string: list of dictionaries}
        Each entry outlines parameters responsible for flows between subpopulations.
    all_states_index : dictionary
        Keys are all the states values are the associated indexes for use with numpy.arrays.
    state_index : dictionary {tuple of strings or ints: {string : int}}
        First level: keys are the subpopulation coordinates are another dictionary.
            Second level: keys are the states and values (ints) are the associated indexes for use with
                          numpy.arrays.
    infected_states_index_list : list of ints
        A list of the indexes of infected states.
    infectious_symptomatic_indexes : dictionary {tuple of strings or ints: [int]}
        Keys are the subpopulation coordinates are a list of indexes for infectious and symptomatic states.
    infectious_asymptomatic_indexes : dictionary {tuple of strings or ints: [int]}
        Keys are the subpopulation coordinates are a list of indexes for infectious and asymptomatic states.
    infectious_and_symptomatic_states : list of stings
        A list of infectious and symptomatic states.
    infectious_and_asymptomatic_states : list of stings
        A list of infectious and asymptomatic states.
    total_states : int
        Total number of states in model.
    subpop_coordinates : list [tuples of ints/strings]
        Coordinates of all subpopulations.
    subpop_suffixes : list of stings
        Suffixes to appended to parameters specific to subpopulations.
    _foi_population_focus : string or None (default None)
        If None a subpopulations force of infection is not divided by  the total population of a subpopulation.
        If 'i' a subpopulations force of infection is divided by the total population of subpopulation 'i',
        i.e. its own population (the population that is being transmitted to).
        If 'j' a subpopulations force of infection for each interacting subpopulation is divided by the total
        population of subpopulation 'j', i.e. the subpopulation it is interacting with (the population that is being
        transmitted from).

    Methods
    -------
    set_structure(self, scaffold, foi_population_focus=None)
        Set metapopulation structure using scaffold.
    ode(self, y, t, *parameters):
        Evaluate the ODE given states (y), time (t) and parameters
    integrate(x0, t, full_output=False, **kwargs_to_pass_to_odeint)
        Simulate model via integration using initial values x0 of time range t.
        A wrapper on top of :mod:`odeint <scipy.integrate.odeint>` giving ode method to odeint.
    get_state_index_dict_of_coordinate(self, coordinate, axis=0):
        Fetch state index dictionaries for given coordinate on an axis.
    get_indexes_of_coordinate(self, coordinate, axis=0):
        Fetch a list of indices for all states for given coordinate on axis.

    Notes
    -----
    I have tried to use notation find in:
        Keeling, M. J., & Rohani, P. (2008). Metapopulations. In Modeling Infectious Diseases in Humans and Animals
        (pp. 237–240). Princeton University Press.
    """
    states = None
    infected_states = None
    infectious_states = infected_states
    symptomatic_states = infected_states
    observed_states = None
    transmission_term_prefix = 'beta'
    subpop_interaction_prefix = 'rho'
    population_term_prefix = 'N'
    _foi_population_focus = None
    asymptomatic_transmission_modifier = None
    universal_params = None
    _dimensions = []
    subpop_params = None  # does not include transmission term beta.
    _subpop_model = None
    len = 0

    def __init__(self, dimensions, subpop_model=None, **other_model_attributes):
        if other_model_attributes:
            for name, value in other_model_attributes.items():
                if not hasattr(self, name):
                    raise AssertionError(name + 'is not a model attribute.')
                setattr(self, name, value)

        if self.states is None:
            raise AssertionError('Model attribute "states" needs to be defined at initialisation of MetaCaster,' +
                                 ' or given as an attribute of a child class of MetaCaster.')

        if self.universal_params is None and self.subpop_params is None:
            raise AssertionError('At least one of the model attributes "universal_params" or "subpop_params" ' +
                                 'needs to be defined at initialisation of MetaCaster,' +
                                 ' or given as an attribute of a child class of MetaCaster.')

        for model_attribute in ['states',
                                'infected_states',
                                'infectious_states',
                                'symptomatic_states',
                                'observed_states',
                                'universal_params',
                                'subpop_params']:
            if eval('self.' + model_attribute) is not None:
                if not _is_set_like_of_strings(eval('self.' + model_attribute)):
                    raise TypeError('Model attribute "' +
                                    model_attribute +
                                    '" should be a collection of unique strings.')

        if subpop_model is not None:
            self.subpop_model = subpop_model

        if type(self) == MetaCaster and other_model_attributes==False and subpop_model is None:
            raise AssertionError('MetaCaster is not meant to run models without other_model_attributes and subpop_model.' +
                                 ' Child classes of MetaCaster can if coded with other_model_attributes and subpop_model.' +
                                 '\nIf unfamiliar with class inheritance look  up:\n' +
                                 ' https://www.w3schools.com/python/python_inheritance.asp.')
        self.dimensions = dimensions

    def subpop_transfer(self, y, y_deltas, t, from_coordinates, parameters):
        """
        Calculates the transfers of people from a subpopulation to all other subpopulations.

        Parameters
        ----------
        y : numpy.array
            Values of variables at time t.
        y_deltas : numpy.Array
            Store of delta (derivative at t) of variables in y which this method adds/subtracts to.
        t : float
            Time t for which derivative is being calculated.
        from_coordinates : tuple or list of strings
            Coordinates for subpopulation being transferred.
        parameters : dictionary {keys are strings: values are numeric}
            Dictionary of parameter values used in calculating derivative.

        Returns
        -------
        y_deltas : numpy.array
            Store of delta (derivative) of variables in y which this method adds/subtracts to.

        """
        if from_coordinates in self.subpop_transfer_dict:
            from_state_index_dict = self.state_index[from_coordinates]
            for group_transfer in self.subpop_transfer_dict[from_coordinates]:
                parameter = group_transfer['parameter']
                if 'piecewise targets' in group_transfer:
                    # This section allows for the piecewise estimation of a people being transferred between groups.
                    if t in self.piecewise_est_param_values[parameter]:
                        param_val = self.piecewise_est_param_values[parameter][t]
                    else:
                        index_of_t = int(t) + 1
                        total_being_tranfered = group_transfer['piecewise targets'][index_of_t]
                        if total_being_tranfered == 0:  # No point in calculations if no one is being vaccinated.
                            param_val = 0
                        else:
                            from_states_index = [from_state_index_dict[state] for state in group_transfer['states']]
                            total_avialable = y[from_states_index].sum()
                            param_val = self._instantaneous_transfer(total_being_tranfered,
                                                                     total_avialable, t)

                        self.piecewise_est_param_values[parameter][t] = param_val
                else:
                    param_val = parameters[parameter]

                to_coordinates = group_transfer['to_coordinates']
                to_state_index_dict = self.state_index[to_coordinates]
                for state in group_transfer['states']:
                    from_index = from_state_index_dict[state]
                    to_index = to_state_index_dict[state]
                    transferring = param_val * y[from_index]
                    y_deltas[from_index] -= transferring
                    y_deltas[to_index] += transferring

        return y_deltas

    def ode(self, y, t, *parameters):
        """
        Evaluate the ODE given states (y), time (t) and parameters


        Parameters
        ----------
        y : numpy.array
            State variables.
        t : float
            Time.
        parameters : floats
            Parameter values.

        Returns
        -------
        y_delta: `numpy.ndarray`
            Deltas of state variables at time point t.
        """
        if self.subpop_model is None:
            raise AssertionError('subpop_model needs to set before simulations can run.')

        subpop_model = self.subpop_model
        subpop_model_arg_names = getfullargspec(subpop_model)[0]
        parameters = dict(zip(self.non_piece_wise_params_names, parameters))
        if self.infectious_states is not None:
            fois = self.calculate_fois(y, parameters, t)
        y_deltas = np.zeros(self.total_states)
        for coordinates in self.subpop_coordinates:
            y_deltas = self.subpop_transfer(y, y_deltas, t, coordinates, parameters)
            if self.infectious_states is not None:
                foi = fois[coordinates]  # force of infection experienced by this specific cluster.

            subpop_suffix = self.coordinates_to_subpop_suffix(coordinates)
            subpop_model_args = {'y': y,
                                 'y_deltas': y_deltas,
                                 'coordinates': coordinates,
                                 'subpop_suffix': subpop_suffix,
                                 'parameters': parameters,
                                 'states_index': self.state_index[coordinates],
                                 't': t,
                                 'foi': foi
                                 }
            subpop_model_args = select_dict_items_in_list(subpop_model_args, subpop_model_arg_names)
            y_deltas = subpop_model(**subpop_model_args)

        return y_deltas

    def calculate_fois(self, y, parameters, t):
        """
        Calculates the Forces of Infection (FOIs) experienced by each subpopulation.

        Parameters
        ----------
        y : numpy.array
            Values of state variables at current time.
        parameters : dictionary {strings: numeric}
            Dictionary of parameter values used in calculating derivative.
        t : float
            Time.

        Returns
        -------
        fois : dictionary {tupple of strings or ints: float}
            Dictionary of the FOIs experienced at each coordinate.

        """
        if self.asymptomatic_transmission_modifier is not None:
            asymptomatic_transmission_modifier = parameters[self.asymptomatic_transmission_modifier]
        else:
            asymptomatic_transmission_modifier = 1

        fois = {coordinates: 0 for coordinates in self.subpop_coordinates}
        for coordinates_i in self.subpop_coordinates:
            subpop_prefix_i = self.coordinates_to_subpop_suffix(coordinates_i)
            beta = parameters[self.transmission_term_prefix + subpop_prefix_i]
            interactions_with_infectious = [self._subpop_infected_interaction(y,
                                                                              coordinates_i,
                                                                              coordinates_j,
                                                                              parameters,
                                                                              asymptomatic_transmission_modifier,
                                                                              t)
                                            for coordinates_j in self.subpop_coordinates]

            interactions_with_infectious = sum(interactions_with_infectious)
            fois[coordinates_i] += beta * interactions_with_infectious

        return fois

    def _subpop_infected_interaction(self, y, coordinates_i, coordinates_j, parameters,
                                     asymptomatic_transmission_modifier, t):
        total_asymptomatic = (asymptomatic_transmission_modifier *
                              y[self.infectious_asymptomatic_indexes[coordinates_j]].sum())
        total_symptomatic = y[self.infectious_symptomatic_indexes[coordinates_j]].sum()
        subpop_prefix_i = self.coordinates_to_subpop_suffix(coordinates_i)
        subpop_prefix_j = self.coordinates_to_subpop_suffix(coordinates_j)
        contribution = parameters[self.subpop_interaction_prefix + subpop_prefix_i + subpop_prefix_j] \
                       * sum([total_asymptomatic, total_symptomatic])
        if self._foi_population_focus == 'i':
            contactable_population = parameters[self.population_term_prefix + '_[' + coordinates_i + ']']
            if callable(contactable_population):
                contactable_population = contactable_population(model=self,
                                                                y=y,
                                                                coordinates=coordinates_i,
                                                                parameters=parameters,
                                                                t=t)
            contribution = contribution / contactable_population

        if self._foi_population_focus == 'j':
            contactable_population = parameters[self.population_term_prefix + '_[' + coordinates_j + ']']
            if callable(contactable_population):
                contactable_population = contactable_population(model=self,
                                                                y=y,
                                                                coordinates=coordinates_j,
                                                                parameters=parameters,
                                                                t=t)
            contribution = contribution / contactable_population

        return contribution

    def get_state_index_dict_of_coordinate(self, coordinate, axis=0):
        """
        Fetch state index dictionaries for given coordinate on an axis.

        Parameters
        ----------
        coordinate : int or str
            Coordinate of
        axis : int default = 0
            Axis on which coordinate is found.

        Returns
        -------
        nested dict: {tuple of int/str: {str: int}}
        """
        selected_coordinates = [coordinates
                                for coordinates in self.subpop_coordinates
                                if coordinates[axis] == coordinate]
        return {coordinates: sub_dict
                for coordinates, sub_dict in self.state_index.items()
                if coordinates in selected_coordinates}

    def get_indexes_of_coordinate(self, coordinate, axis=0):
        """
        Fetch a list of indices for all states for given coordinate on axis.

        Parameters
        ----------
        coordinate : int or str
            Coordinate of
        axis : int default = 0
            Axis on which coordinate is found.

        Returns
        -------
        list of ints
        """
        selected_state_indexes = self.get_state_index_dict_of_coordinate(self, coordinate, axis=axis)
        return _nested_dict_values(selected_state_indexes)

    @staticmethod
    def coordinates_to_subpop_suffix(coordinates):
        """
        Transform coordinates to subpopulation suffix appended to subpopulation specific parameters.

        Parameters
        ----------
        coordinates : list/tuple of ints or strings
            Coordinate of subpopulation

        Returns
        -------
        string
        """
        if isinstance(coordinates, (list, tuple)):
            if all(isinstance(coordinate, int) for coordinate in coordinates):
                coordinates = (str(coordinate) for coordinate in coordinates)
            if any(not isinstance(coordinate, str) for coordinate in coordinates):
                raise TypeError('All coordinates must be either string or integers, or lists/tuples of those types.')
            return '_[' + ','.join(coordinates) + ']'
        elif isinstance(coordinates, str):
            return '_[' + coordinates + ']'
        elif isinstance(coordinates, int):
            return '_[' + str(coordinates) + ']'
        else:
            raise TypeError('All coordinates must be either string or integers, or lists/tuples of those types.')

    def _instantaneous_transfer(self, population_transitioning, population, t=None):
        """
        Calculate instantaneous rate needed to reduce population by population transitioning to another compartment.

        Parameters
        ----------
        population_transitioning : float
            Population moving between compartments.
        population : float
            Population from which transition is taking place.
        t : float
            Time at t. Used in generating an error specific to simulating models using this class.

        Returns
        -------
        float
            Instantaneous rate of change for the time t.
        """
        if population_transitioning > population:
            error_msg = "population_transitioning (" + str(
                population_transitioning) + ") is greater than population (" + str(population)
            if t is None:
                error_msg += ').'
            else:
                error_msg += '), at time ' + str(t) + ').'
            raise ValueError(error_msg)
        if population_transitioning == 0 and population == 0:
            return 0
        else:
            proportion_by_t = population_transitioning / population
            return -np.log(1 - proportion_by_t)

    def integrate(self, x0, t, full_output=False, **kwargs_to_pass_to_odeint):
        """
        Simulate model via integration using initial values x0 of time range t.

        A wrapper on top of :mod:`odeint <scipy.integrate.odeint>`, giving ode method to odeint.
        Modified method from the pygom method `DeterministicOde <pygom.model.DeterministicOde>`.
        
        Parameters
        ----------
        x0 : array like
            Initial values of states.
        t : array like
            Timeframe over which model is to be simulated.
        full_output : bool, optional
            If additional information from the integration is required
        kwargs_to_pass_to_odeint : dictionary
            Key word arguments to pass to scipy.integrate.odeint.

        Returns
        -------
        solution: pandas.DataFrame
            Multi-index columns are  clusters by vaccine groups by states.
        """
        self.piecewise_est_param_values = {param: {} for param in self.params_estimated_via_piecewise_method}
        # INTEGRATE!!! (shout it out loud, in Dalek voice)
        # determine the number of output we want
        args = tuple(self.parameters.values())
        # The if else statement below is for use with DOK matrix version of the Jacobian see sileneced section below.
        # if self.dok_jacobian is None: # May or may not of defined the models Jacobian
        #     solution, output = scipy.integrate.odeint(self.ode,
        #                                               x0, t, args=args,
        #                                               full_output=True,
        #                                               **kwargs_to_pass_to_odeint)
        # else:
        #     solution, output = scipy.integrate.odeint(self.ode,
        #                                               x0, t, args=args,
        #                                               Dfun=self.jacobian,
        #                                               full_output=True,
        #                                               **kwargs_to_pass_to_odeint)
        solution, output = scipy.integrate.odeint(self.ode,
                                                  x0, t, args=args,
                                                  full_output=True,
                                                  **kwargs_to_pass_to_odeint)
        solution = self.results_array_to_df(solution, t)
        if full_output:
            # have both
            return solution, output
        else:
            return solution

    def results_array_to_df(self, results, t):
        """
        Converts array of results into a dataframe with multi-index columns reflecting meta-population structure.

        Parameters
        ----------
        results : np.array
            Results from simulations of model.
        t : np.array
            Time over which model was simulated.

        Returns
        -------
        results_df : pandas.DataFrame
            Results with multi-index columns reflecting meta-population structure.

        """
        state_index = self.state_index
        if len(self) == 1:
            multi_columns = [(coordinates, state)
                             for coordinates, sub_dict in self.state_index.items()
                             for state in sub_dict.keys()
                             ]
        else:
            multi_columns = [(','.join(coordinates), state)
                             for coordinates, sub_dict in self.state_index.items()
                             for state in sub_dict.keys()
                             if coordinates != 'observed_states'
                             ]
            multi_columns += [('observed_states', state)
                              for state in self.state_index['observed_states'].keys()
                              ]
        results_df = pd.DataFrame(results, index=t)
        results_df.columns = pd.MultiIndex.from_tuples(multi_columns)
        return results_df

    def __len__(self):
        """
        Number of axis/dimensions of the metapopulation model.

        Returns
        -------
        int
        """
        return len(self._dimensions)

    @property
    def dimensions(self):
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions):
        """
        Set metapopulation structure using scaffold.

        Parameters
        ----------
        dimensions : int, collection of unique strings, list/tuple of ints, list/tuple of collections of unique strings OR list/tuple of dictionaries (transfer dictionaries)
            If int :
                This creates a one dimension metapopulation structure with range(scaffold) used to label subpopulations
                in dimensions attribute.
            If list/tuple/set of unique strings:
                This creates a one dimension metapopulation structure with entries used to label subpopulations
                 in dimensions attribute.
            If list/tuple of ints:
                 This creates a multidimensional metapopulation structure with range(each int entry) used to generate
                  labels on an axis of the subpopulations in dimensions attribute.
            If list/tuple of list/tuple/set of unique strings:
             This creates a multidimensional metapopulation structure with each sub-list/tuple/set entries used as
              labels in am axis of dimensions attribute.
            If  list/tuple of dictionaries (transfer dictionaries):
                Transfer dictionaries outlines the transfer of one subpopulation to another subpopulation.
                Each transfer dictionary must have the key values pairs:
                    from_coordinates: string/int or list/tuple of strings/ints
                        Subpopulation coordinates from which hosts are leaving. All of these entries should be of the same
                         length.
                    to_coordinates: string/int or list/tuple of strings/ints
                        Subpopulation coordinates from which hosts are leaving.All of these entries should be of the same
                         length and the same length as the from_coordinates entries.
                    states: list of strings or string
                        Host states which will transition between subpopulations. Single entry of 'all' value
                        means all the available model states transition between subpopulations. Alternatively a list of
                        specific states can be given.
                    parameter : string
                        Name given to parameter that is responsible for flow of hosts transferring between subpopulations.
                Optional key value pairs:
                    piecewise targets: list, tuple, numpy.Array or pandas.Series
                        Targets for piecewise estimation of parameter that is responsible for flow of hosts transitions
                        between clusters and vaccine groups (see method group_transfer).

        Returns
        -------
        None

        Notes: Attributes Altered
        -------------------------
        parameter_names : list of strings
            The names of all the parameters.
        population_terms : list of strings
            Subpopulation population terms used in calculating force of infection (see method calculate_fois).
        transmission_terms :  list of strings
            The transmission terms for each subpopulation (see method calculate_fois).
        subpop_interaction_terms : list of strings
            The terms for each interaction between subpopulations used when calculating force of infection
            (see method calculate_fois).
        total_subpops : int
            Total number of subpopulations.
        total_parameters : int
            Total number of parameters in model.
        dimensions : list of sets of ints/strings
            Dimensions of metapopulation.
        subpop_transfer_dict : dictionary {tuple of ints or strings: list of dictionaries}
            Each entry outlines all the outflows from a subpopulation to other subpopulation. For use with
            subpop_transfer method.
        subpop_transition_params_dict : dictionary {string: list of dictionaries}
            Each entry outlines parameters responsible for flows between subpopulations.
        all_states_index : dictionary
            Keys are all the states values are the associated indexes for use with numpy.arrays.
        state_index : dictionary {tuple of strings or ints: {string : int}}
            First level: keys are the subpopulation coordinates are another dictionary.
                Second level: keys are the states and values (ints) are the associated indexes for use with
                              numpy.arrays.
        infected_states_index_list : list of ints
            A list of the indexes of infected states.
        infectious_symptomatic_indexes : dictionary {tuple of strings or ints: [int]}
            Keys are the subpopulation coordinates are a list of indexes for infectious and symptomatic states.
        infectious_asymptomatic_indexes : dictionary {tuple of strings or ints: [int]}
            Keys are the subpopulation coordinates are a list of indexes for infectious and asymptomatic states.
        infectious_and_symptomatic_states : list of stings
            A list of infectious and symptomatic states.
        infectious_and_asymptomatic_states : list of stings
            A list of infectious and asymptomatic states.
        total_states : int
            Total number of states in model.
        subpop_coordinates : list [tuples of ints/strings]
            Coordinates of all subpopulations.
        subpop_suffixes : list of stings
            Suffixes to appended to parameters specific to subpopulations.
        """
        self.parameter_names = set(self.universal_params)
        if self.asymptomatic_transmission_modifier is not None:
            self.parameter_names.add(self.asymptomatic_transmission_modifier)

        #### Using Scaffold to define dimensions ####
        self.params_estimated_via_piecewise_method = []
        self.subpop_transfer_dict = {}
        self.subpop_transition_params_dict = {}
        if isinstance(dimensions, (list, tuple)) and all(isinstance(item, dict) for item in dimensions):
            for count, group_transfer in enumerate(dimensions):
                if 'from_coordinates' not in group_transfer:
                    raise ValueError(
                        'If scaffold is a list of subpopulation transfer dictionaries it must have a "from_coordinates"' +
                        ' entry in every subpopulation transfer dictionary,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                from_coordinates = group_transfer['from_coordinates']
                if not isinstance(from_coordinates, (list, tuple)):
                    raise TypeError(
                        'If scaffold is a list of subpopulation transfer dictionaries values for "from_coordinates"' +
                        '  entries should be a list or a tuple,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                if not (all(isinstance(coordinate, int) for coordinate in from_coordinates) or
                        all(isinstance(coordinate, str) for coordinate in from_coordinates)):
                    raise TypeError(
                        'If scaffold is a list of subpopulation transfer dictionaries values for "from_coordinates"' +
                        '  entries should be a list or a tuple of only strings or integers,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                if count == 0:
                    self._dimensions = [{coordinate} for coordinate in from_coordinates]
                elif len(from_coordinates) != len(self):
                    raise ValueError('If scaffold is a list of dictionaries all "from_coordinates"' +
                                     ' entries in subpopulation transfer dictionaries should be of the same length' +
                                     ', check subpopulation transfer dictionary [' +
                                     str(count) +
                                     '] of scaffold.')
                else:
                    for axis, coordinate in enumerate(from_coordinates):
                        self._dimensions[axis].add(coordinate)

                if from_coordinates not in self.subpop_transfer_dict:
                    self.subpop_transfer_dict[from_coordinates] = []

                entry = {}
                if 'to_coordinates' not in group_transfer:
                    raise ValueError(
                        'If scaffold is a list of subpopulation transfer dictionaries it must have a "to_coordinates"' +
                        ' entry in every subpopulation transfer dictionary,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                to_coordinates = group_transfer['to_coordinates']
                if not isinstance(to_coordinates, (list, tuple)):
                    raise TypeError(
                        'If scaffold is a list of subpopulation transfer dictionaries values for "to_coordinates"' +
                        '  entries should be a list or a tuple,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                if not (all(isinstance(coordinate, int) for coordinate in to_coordinates) or
                        all(isinstance(coordinate, str) for coordinate in to_coordinates)):
                    raise TypeError(
                        'If scaffold is a list of subpopulation transfer dictionaries values for "to_coordinates"' +
                        '  entries should be a list or a tuple of only strings or integers,' +
                        ' check subpopulation transfer dictionary [' +
                        str(count) +
                        '] of scaffold.')
                if len(to_coordinates) != len(self):
                    raise ValueError('If scaffold is a list of dictionaries all "to_coordinates" and' +
                                     ' "from_coordinates" entries in subpopulation transfer dictionaries should be of the' +
                                     ' same length' +
                                     ', check "to_coordinates" subpopulation transfer dictionary [' +
                                     str(count) +
                                     '] of scaffold.')
                for axis, coordinate in enumerate(to_coordinates):
                    self._dimensions[axis].add(coordinate)

                entry['to_coordinates'] = to_coordinates
                if group_transfer['states'] == 'all':
                    entry['states'] = self.states
                else:
                    states = copy.deepcopy(group_transfer['states'])
                    if isinstance(states, str):
                        states = list(states)
                    if not _is_set_like_of_strings(states):
                        raise ValueError('"states" entry in Subpopulation transfer dictionary [' +
                                         str(count) +
                                         '] of scaffold is not a collection of unique items.')
                    for state in states:
                        if state not in self.states:
                            raise ValueError('Subpopulation transfer dictionary [' +
                                             str(count) +
                                             '] of scaffold has an unrecognised state (' + state + ').' +
                                             'All states in listed in a scaffold should be one of those listed in the' +
                                             ' subpopulation model.')
                    entry['states'] = states

                parameter = group_transfer['parameter']
                if not isinstance(parameter, str):
                    raise TypeError('In subpopulation transfer dictionary [' + str(count) +
                                    '] of scaffold ' + str(parameter) + ' should be of type string.')
                entry['parameter'] = parameter
                if parameter not in self.subpop_transition_params_dict:
                    self.subpop_transition_params_dict[parameter] = []
                self.subpop_transition_params_dict[parameter].append({key: value for key, value in
                                                                      group_transfer.items()
                                                                      if key != 'parameter'})
                self.parameter_names.add(parameter)
                if 'piecewise targets' in group_transfer:
                    self.params_estimated_via_piecewise_method.append(parameter)
                    if isinstance(group_transfer['piecewise targets'], pd.Series):
                        entry['piecewise targets'] = group_transfer['piecewise targets'].to_numpy()
                    elif isinstance(group_transfer['piecewise targets'], (list, tuple)):
                        entry['piecewise targets'] = np.array(group_transfer['piecewise targets'])
                    elif isinstance(group_transfer['piecewise targets'], np.ndarray):
                        entry['piecewise targets'] = group_transfer['piecewise targets']

                accepted_keys = list(entry.keys()) + ['from_coordinates']
                for key in group_transfer.keys():
                    if key not in accepted_keys:
                        raise ValueError('Subpopulation transfer dictionary [' +
                                         str(count) +
                                         '] of scaffold has an unrecognised entry (' + key + ').')

                self.subpop_transfer_dict[from_coordinates].append(entry)
        elif (isinstance(dimensions, (list, tuple)) and
              all(_is_set_like_of_strings(item) for item in dimensions)):
            self._dimensions = [set(item) for item in dimensions]
        elif (isinstance(dimensions, (list, tuple)) and
              all(isinstance(item, int) for item in dimensions)):
            self._dimensions = [set(*range(num)) for num in dimensions]
        elif isinstance(dimensions, (list, tuple)) and _is_set_like_of_strings(dimensions):
            self._dimensions = [set(dimensions)]
        elif isinstance(dimensions, set) and all(isinstance(item, str) for item in dimensions):
            self._dimensions = [dimensions]
        elif isinstance(dimensions, int):
            self._dimensions = [set(*range(int))]
        else:
            raise TypeError('scaffold is not supported.')

        #### Definining States based on new Dimensions ####
        self.infectious_and_symptomatic_states = [state for state in self.infectious_states
                                                  if state in self.symptomatic_states]
        self.infectious_and_asymptomatic_states = [state for state in self.infectious_states
                                                   if state not in self.symptomatic_states]
        self.all_states_index = {}
        self.state_index = {}
        self.infectious_symptomatic_indexes = {}
        self.infectious_asymptomatic_indexes = {}
        self.infected_states_index_list = []
        # populating index dictionaries
        index = 0
        self.subpop_coordinates = []
        self.subpop_suffixes = []
        if len(self) == 1:
            axis_equals_1 = True
        else:
            axis_equals_1 = False
        for coordinates in itertools.product(*self._dimensions):
            subpop_suffix = self.coordinates_to_subpop_suffix(coordinates)
            self.subpop_suffixes.append(subpop_suffix)
            if axis_equals_1:
                coordinates = coordinates[0]
            self.subpop_coordinates.append(coordinates)
            self.state_index[coordinates] = {}
            self.infectious_symptomatic_indexes[coordinates] = []
            self.infectious_asymptomatic_indexes[coordinates] = []
            for state in self.states:
                all_state_index_key = state + subpop_suffix
                self.all_states_index[all_state_index_key] = index
                self.state_index[coordinates][state] = index
                if state in self.infectious_and_symptomatic_states:
                    self.infectious_symptomatic_indexes[coordinates].append(index)
                if state in self.infectious_and_asymptomatic_states:
                    self.infectious_asymptomatic_indexes[coordinates].append(index)
                if state in self.infected_states:
                    self.infected_states_index_list.append(index)
                index += 1

        self.state_index['observed_states'] = {}
        for state in self.observed_states:
            self.all_states_index[state] = index
            self.state_index['observed_states'][state] = index
            index += 1

        self.total_states = index
        for transfer_info in self.subpop_transition_params_dict.values():
            for transfer_info_entry in transfer_info:
                from_coordinates = transfer_info_entry['from_coordinates']
                from_states_dict = self.state_index[from_coordinates]
                to_coordinates = transfer_info_entry['to_coordinates']
                to_states_dict = self.state_index[to_coordinates]
                state_selection = transfer_info_entry['states']
                if state_selection == 'all':
                    transfer_info_entry['from_index'] = [from_states_dict.values()]
                    transfer_info_entry['to_index'] = [to_states_dict.values()]
                else:
                    transfer_info_entry['from_index'] = [from_states_dict[state] for state in state_selection]
                    transfer_info_entry['to_index'] = [to_states_dict[state] for state in state_selection]

        #### Defining Parameters based on new Dimensions ####
        self.parameter_names.update([param + subpop_suffix
                                     for param in self.subpop_params
                                     for subpop_suffix in self.subpop_suffixes
                                     ])

        self.transmission_terms = [
            self.transmission_term_prefix + subpop_suffix
            for subpop_suffix in self.subpop_suffixes]
        self.subpop_interaction_terms = [
            self.subpop_interaction_prefix + subpop_suffix_i + subpop_suffix_j
            for subpop_suffix_i in self.subpop_suffixes
            for subpop_suffix_j in self.subpop_suffixes
        ]
        self.parameter_names.update(self.subpop_interaction_terms)
        if self._foi_population_focus is not None:
            self.population_terms = [self.population_term_prefix + subpop_suffix
                                     for subpop_suffix in self.subpop_suffixes]
            self.parameter_names.update(self.population_terms)

        self.parameter_names.update(self.transmission_terms)
        self.total_subpops = len(self.subpop_coordinates)
        self.parameter_names = sorted(self.parameter_names)
        non_piece_wise_params_names = set(self.parameter_names) - set(self.params_estimated_via_piecewise_method)
        self.non_piece_wise_params_names = sorted(list(non_piece_wise_params_names))
        self._parameters = None
        self.total_parameters = len(self.parameter_names)
        self.piecewise_est_param_values = None

    @property
    def foi_population_focus(self):
        return self._foi_population_focus

    @foi_population_focus.setter
    def foi_population_focus(self, foi_population_focus):
        if foi_population_focus is not None:
            if foi_population_focus not in ['i', 'j']:
                raise ('population_denominator_in_foi can only be change to "i" or "j" or None.')

        self._foi_population_focus = foi_population_focus

    @property
    def subpop_model(self):
        return self._subpop_model

    @subpop_model.setter
    def subpop_model(self, subpop_model):
        if not callable(subpop_model):
            raise TypeError('subpop_model should be a callable function.')
        subpop_model_arg_names = getfullargspec(subpop_model)[0]
        required_args = ['y', 'y_deltas', 'parameters', 'states_index']
        for arg in required_args:
            if arg not in subpop_model_arg_names:
                raise ValueError(arg + ' is missing from subpop_model function.')

        expected_args = ['y', 'y_deltas', 'parameters', 'states_index', 'coordinates', 'subpop_suffix', 'foi', 't']
        for arg in subpop_model_arg_names:
            if arg not in expected_args:
                raise ValueError(arg + ' in not a supported argument for the subpop_model function. ' +
                                 'Supported arguments include ' + ', '.join(expected_args))
        self._subpop_model = subpop_model

    @property
    def parameters(self):
        """
        Returns
        -------
        Dictionary
            A dictionary of parameter values.

        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """
        Set parameter values for simulations.

        Parameters
        ----------
        parameters : dictionary {key: Numeric or callable}
            Parameter values. In the case of population_terms these can be callable function/class methods
            which calculate populations at a given time point. If so population_terms must have the arguments 'model',
            'y', 'parameters', 'coordinates', 't'.

        Returns
        -------
        Nothing
        """
        if not isinstance(parameters, dict):
            raise TypeError('Currently non non_piecewise_params must be entered as a dict.')
        # we assume that the key of the dictionary is a string and
        # the value can be a single value or a distribution

        for param_name, value in parameters.items():
            if param_name not in self.parameter_names:
                raise ValueError(param_name + ' is not a name given to a parameter for this model.')
            if param_name in self.params_estimated_via_piecewise_method:
                raise AssertionError(param_name + ' was set as a parameter to be estimated via piecewise estimiation ' +
                                     'at the initialization of this model.')
            if callable(value):
                function_arg_names = getfullargspec(value)[0]
                if any(args not in ['model', 'y', 'parameters', 'coordinates', 't'] for args in function_arg_names):
                    raise ValueError(param_name +
                                     " is a function but not a population_term but does not have all of the arguments" +
                                     " 'model', 'y', 'parameters', 'coordinates' or 't'.")
            elif not isinstance(value, Number):
                raise TypeError(param_name + ' should be a number type.')
        params_not_given = [param for param in self.parameter_names
                            if param not in
                            list(parameters.keys()) + self.params_estimated_via_piecewise_method]
        if params_not_given:
            raise Exception(', '.join(params_not_given) +
                            " are/is missing from parameters for model (see self.all_parameters).")
        # this must be sorted alphanumerically.
        self._parameters = {key: value for key, value in sorted(parameters.items())}

    @property
    def shape(self):
        """
        Number of entries in each axis of the metapopulation model.

        Returns
        -------
        tuple of ints
        """
        return (len(axis) for axis in self._dimensions)


if __name__ == "__main__":
    pass



"""
Creation:
    Author: Martin Grunnill
    Date: 2024/03/08
Description: Event que for simulating models between events.
    
"""
from .events import (
    BaseEvent,
    TransferEvent,
    ChangeParametersEvent,
    ParametersEqualSubPopEvent,
    )
from collections import OrderedDict, deque
from numbers import Number
import numpy as np
import pandas as pd
from warnings import warn
import copy
import inspect



class EventQueue:
    """
    An event queue for processing events from lowest to the highest time. User defined method simulates model between events.

    See method run_simulation on how to customise simulating of model between events.

    Parameters
    ----------
    events : list(either event_handling.BaseEvent or subclass of event_handling.BaseEvent)
        Events to go into event queue.

    Methods
    -------
    reset_event_queue()
        Resets event queue to state at initialisation.
    change_event_proportion(event_names, proportion)
        Change proportion attribute value of events in event_names.
    change_event_factor(event_names, factor)
        Change factor attribute value of events in event_names.
    change_event_value(event_names, value)
        Change value attribute value of events in event_names.
    make_events_nullevents(event_names)
        Turns events into nullevents (do nothing events).
    get_event_names()
        Return list of event names.
    events_at_same_time()
        Returns a dictionary of events occurring at same time. Keys are time values are lists of event names.
    run_simulation(model_object, run_attribute, y0, end_time, parameters_attribute, parameters,
                       start_time=0, simulation_step=1,
                       full_output=False, return_param_changes=False,
                       **kwargs_to_pass_to_func)
        Carries out events in queue, running simulations using user defined method between events.


    """

    def __init__(self, events):
        self._event_queue = _EventQueue(events)
        self._master_event_queue = copy.deepcopy(self._event_queue)
        self._events = self._event_queue._events

    def reset_event_queue(self):
        """
        Resets event queue to state at initialisation.

        Returns
        -------
        Nothing
        """
        self._event_queue = copy.deepcopy(self._master_event_queue)
        self._events = self._event_queue._events

    def _event_names_checker(self, event_names):
        if event_names == 'all':
            event_names = self.get_event_names()
        else:
            if not pd.api.types.is_list_like(event_names):
                event_names = [event_names]
            if any(not isinstance(item, str) for item in event_names):
                raise TypeError('All event_names entries must be a string.')
            available_event_names = self.get_event_names()
            for event_name in event_names:
                if event_name not in available_event_names:
                    raise TypeError(event_name + ' not listed in names give to available evednt: ' +
                                    ','.join(available_event_names) +  '.')
        return event_names

    def change_event_proportion(self, event_names, proportion):
        """
        Change proportion attribute value of events in event_names.

        Parameters
        ----------
        event_names : list-like of strings or single string
            Name of event(s). If 'all' is entered attribute is changed for all events.
        proportion : float/int between 0 and 1
            Value for which proportion attribute is changed to.

        Returns
        -------
        Nothing
        """
        event_names = self._event_names_checker(event_names)
        for event_name in event_names:
            event = self._events[event_name]
            event.proportion = proportion

    def change_event_factor(self, event_names, factor):
        """
        Change factor attribute value of events in event_names.

        Parameters
        ----------
        event_names : list-like of strings or single string
            Name of event(s). If 'all' is entered attribute is changed for all events.
        factor : int/float
            Value for which factor attribute is changed to.

        Returns
        -------
        Nothing
        """
        event_names = self._event_names_checker(event_names)
        for event_name in event_names:
            event = self._events[event_name]
            event.factor = factor

    def change_event_value(self, event_names, value):
        """
        Change value attribute value of events in event_names.

        Parameters
        ----------
        event_names : list-like of strings or single string
            Name of event(s). If 'all' is entered attribute is changed for all events.
        value : int/float
            Value for which value attribute is changed to.

        Returns
        -------
        Nothing
        """
        event_names = self._event_names_checker(event_names)
        for event_name in event_names:
            event = self._events[event_name]
            event.value = value

    def make_events_nullevents(self, event_names):
        """
        Turns events into nullevents (do nothing events).

        Parameters
        ----------
        event_names : list-like of strings or single string
            Name of event(s). If 'all' is entered all events are made nullevents.

        Returns
        -------
        Nothing
        """
        event_names = self._event_names_checker(event_names)
        for event_name in event_names:
            event = self._events[event_name]
            event.make_event_a_nullevent()

    def get_event_names(self):
        """
        Return list of event names.
        Returns
        -------
        List of strings
            Event names.
        """
        return list(self._events.keys())

    def events_at_same_time(self):
        """
        Returns a dictionary of events occurring at same time. Keys are time values are lists of event names.

        Returns
        -------
        dictionary
             Keys are floats/ints values are lists of strings.
        """
        return self._event_queue.events_at_same_time

    def run_simulation(self, model_object, run_attribute, y0, end_time, parameters_attribute, parameters,
                       start_time=0, simulation_step=1,
                       full_output=False, return_param_changes=False,
                       **kwargs_to_pass_to_func):
        """
        Carries out events in queue, running simulations using user defined method between events.

        Parameters
        ----------
        model_object : object
            Object used to define and simulate model.
        run_attribute : string
            Name of model_objects method that simulates model. Must return either a numpy array of pandas.DataFrame.
        y0 : numpy.array
            Intial values of state varibles.
        end_time : float/int
            End time of simulations.
        parameters_attribute : string
            Attribute of model_object that sets parameters (must accept dictionary where keys are strings and values are
            floats/ints).
        parameters : dict {str : floats/ints}
            Parameters being used by model.
        start_time : float/int, default 0
            Start time of simulations.
        simulation_step : float/int, default 1
            Time steps used in simulations.
        full_output : bool, default False
            If true an info_dict is returned outlining full_output information given from running of model between
            events.
        return_param_changes : bool, default is False
            If true dictionary outlining parameters value changes made by events.
        kwargs_to_pass_to_func : **kwargs
            Key word arguments to pass to run_attribute method.

        Returns
        -------
        y : numpy.array or pandas.Dataframe
            Solution from simulating model and processing events.
        transfers_df : pandas.Dataframe
            Details of values being transferred by events between state variables.
        """
        if not all(time % simulation_step == 0
                   for time in self._event_queue.times):
            raise ValueError('All time points for events must be divisible by simulation_step, leaving no remainder.')
        setattr(model_object, parameters_attribute, parameters)
        param_changes = {'Starting value, time: '+str(start_time): copy.deepcopy(parameters)}
        sim_event_queue = copy.deepcopy(self._event_queue)
        sim_event_queue.prep_queue_for_sim_time(start_time, end_time)
        tranfers_list = []
        y = []
        current_time = start_time
        solution_at_t = copy.deepcopy(y0)
        if full_output:
            run_method = getattr(model_object, run_attribute)
            function_args_inspection = inspect.getfullargspec(run_method)
            full_output_in_func_args = 'full_output' in function_args_inspection.args
            if full_output_in_func_args:
                info_dict = {}
            else:
                warn('Full output unavailable as full_output is not an argument in function given to "func".')

            def _func_with_full_output(func, solution_at_t, current_t_range, info_dict, **kwargs_to_pass_to_func):
                # Not sure if this should be a private function outside of class or private class method.
                y_over_current_t_range, info_sub_dict = func(solution_at_t, current_t_range, full_output=True,
                                                             **kwargs_to_pass_to_func)
                range_as_list = current_t_range.tolist()
                time_points = (range_as_list[0], range_as_list[-1])
                info_dict[time_points] = info_sub_dict
                return y_over_current_t_range


        earliest_event_time = self._event_queue.earliest_event_time
        while sim_event_queue.not_empty():
            next_time, event = sim_event_queue.poptop()
            if next_time > end_time: # if the next event is after simulations time break out of loop.
                break

            if current_time != next_time:
                # run until current time is next time
                current_t_range = np.arange(current_time, next_time+simulation_step, simulation_step)
                run_method = getattr(model_object, run_attribute)
                if full_output and full_output_in_func_args:
                    y_over_current_t_range = _func_with_full_output(run_method, solution_at_t, current_t_range,
                                                                    info_dict, **kwargs_to_pass_to_func)
                else:
                    y_over_current_t_range = run_method(solution_at_t, current_t_range, **kwargs_to_pass_to_func)
                if next_time == earliest_event_time:
                    if not isinstance(y_over_current_t_range, (np.ndarray, np.generic, pd.DataFrame)):
                        raise TypeError('run_attribute should return numpy array of pandas dataframe.')
                if isinstance(y_over_current_t_range, (np.ndarray, np.generic)):
                    solution_at_t = y_over_current_t_range[-1, :]
                    y.append(y_over_current_t_range[:-1, :])
                elif isinstance(y_over_current_t_range, pd.DataFrame):
                    solution_at_t = y_over_current_t_range.iloc[-1].to_numpy()
                    y.append(y_over_current_t_range.iloc[:-1])
                current_time = next_time
            # then do event
            if isinstance(event, TransferEvent):
                transfered = event.process(solution_at_t, current_time)
                transfers_entry = {'time':current_time,
                                   'transfered':transfered,
                                   'event':event.name}
                tranfers_list.append(transfers_entry)
            elif isinstance(event, ChangeParametersEvent):
                parameters = event.process(model_object, parameters_attribute, parameters)
                param_changes[event.name + ', time: ' + str(current_time)] = copy.deepcopy(parameters)

        if current_time != end_time:
            current_t_range = np.arange(current_time, end_time+simulation_step, simulation_step)
            run_method = getattr(model_object, run_attribute)
            if full_output and full_output_in_func_args:
                y_over_current_t_range = _func_with_full_output(run_method, solution_at_t, current_t_range, info_dict, **kwargs_to_pass_to_func)
            else:
                y_over_current_t_range = run_method(solution_at_t, current_t_range, **kwargs_to_pass_to_func)
            y.append(y_over_current_t_range)

        if isinstance(y[0], pd.DataFrame):
            y = pd.concat(y)
        else:
            y = np.vstack(y)
        transfers_df = pd.DataFrame(tranfers_list)
        if full_output and full_output_in_func_args:
            if return_param_changes:
                return y, transfers_df, param_changes, info_dict
            else:
                return y, transfers_df, info_dict
        else:
            if return_param_changes:
                return y, transfers_df, param_changes
            else:
                return y, transfers_df


class _EventQueue:
    """
    Event queue internal to EventQueue class above.
    This is kept separate from EventQueue so that a master copy can be kept for reverting changes back to.

    Parameters
    ----------
    events : list(either event_handling.BaseEvent or subclass of event_handling.BaseEvent)
        Events to go into event queue.

    Attributes
    ----------
    queue : OrderedDict
        Queue events ordered by time (keys of ordered dict). When more then one event occurs at a time a deque object
        used to stack events. Events occurring at same time are processed under first one in first one out order.
    events_at_same_time : dict
        A dictionary of events occurring at same time. Keys are time values are lissts of event names.
    times : list of floats or int values.
        Times of events.
    earliest_event_time(self):
        Get the earliest event's time.
    len : int
        Number of events in queue.

    Methods
    -------
    poptop()
        Pops item from top of queue going by order of the lowest event time first.
    prep_queue_for_sim_time(start_time, end_time)
        Edits event queue removing events before start_time or after end_time.
    not_empty()
        Checks that event queue is not empty.

    """

    def __init__(self, events):
        unordered_que = {}
        self.events_at_same_time = {}
        self._events = {}
        if isinstance(events, BaseEvent):
            events = [events]
        if not isinstance(events,(list,tuple,set)):
            raise TypeError('Events should be either a single BaseEvent (or a subclass of BaseEvent) '+
                            'or a list/tuple/set of BaseEvents (or a subclasses of BaseEvent).')
        for event in events:
            self._events[event.name] = event
            times_already_in_queue = set(unordered_que.keys()) & event.times
            if times_already_in_queue:
                for time in times_already_in_queue:
                    if isinstance(unordered_que[time], BaseEvent):
                        self.events_at_same_time[time] = [unordered_que[time].name, event.name]
                        unordered_que[time] = deque([unordered_que[time], event])
                    else:
                        self.events_at_same_time[time].append(event.name)
                        unordered_que[time].append(event)

            times = event.times - times_already_in_queue
            unordered_que.update({time: event for time in times})
        self.queue = OrderedDict(sorted(unordered_que.items()))
        # if self.events_at_same_time:
        #     warn('Concuring events in event queue. To view use method "events_at_same_time".')

    def poptop(self):
        """
        Pops item from top of queue going by order of the lowest event time first.

        Returns
        -------
        event_time : float
            Time of event.
        event : child of event_handling.event.BaseEvent
            An event.
        """
        if self.queue:  # OrderedDicts if not empty are seen as True in bool statements.
            next_item = next(iter(self.queue.values()))
            if isinstance(next_item, deque):
                event_time = next(iter(self.queue.keys()))
                event = next_item.popleft()
                if not next_item:
                    self.queue.popitem(last=False)
                return event_time, event
            else:
                return self.queue.popitem(last=False)
        else:
            raise KeyError("Empty event queue")

    def prep_queue_for_sim_time(self, start_time, end_time):
        """
        Edits event queue removing events before start_time or after end_time.

        Parameters
        ----------
        start_time : float or int
            Start time of simulations.
        end_time : float or int
            End time of simulations.
        Returns
        -------
        Nothing
        """
        first_event_time = next(iter(self.queue.keys()))
        last_event_time = next(reversed(self.queue.keys()))
        if first_event_time < start_time and last_event_time > last_event_time:
            self.queue = OrderedDict({time: item for time, item
                                      in self.queue.items()
                                      if time >= start_time and time <= end_time})
        elif first_event_time < start_time:
            self.queue = OrderedDict({time: item for time, item
                                      in self.queue.items()
                                      if time >= start_time})
        elif last_event_time > last_event_time:
            self.queue = OrderedDict({time: item for time, item
                                      in self.queue.items()
                                      if time <= end_time})

    def not_empty(self):
        """
        Checks that event queue is not empty.
        Returns
        -------
        bool
        """
        return bool(self.queue)

    @property
    def times(self):
        """
        Return a list of the times of events.
        Returns
        -------
        list of floats or int values
            A list of the times of events.
        """
        return list(self.queue.keys())
    @property
    def earliest_event_time(self):
        """
        Get the earliest event's time.

        Returns
        -------
        int or float.
        """
        return self.times[0]

    def __len__(self):
        """
        Returns number of event in queue.

        Returns
        -------
        int
        """
        return len(self.queue)

    def __repr__(self):
        return f"Queue({self.data.items()})"


if __name__ == "__main__":
    pass


"""
Creation:
    Author: Martin Grunnill
    Date: 2024/03/08
Description: Events for use in class EventQueue.

"""
from numbers import Number
from warnings import warn
import numpy as np

class BaseEvent:
    """
    Base parent event.

    Parameters & Attributes
    -----------------------
    name : str
        Name given to event.
    times : floats/ints, ranges or list/tuple/set of floats/ints
        Times at which event occurs.

    Methods
    -------
    process()
        Method for carrying out event.
    make_event_a_nullevent()
        Changes method into a do null event (do nothing event). Child classes process method should start with
        'if self._do_nothing:
            super().process()
        else:'
        or
        'if self._do_nothing:
            pass
        else:'

    undo_make_event_a_nullevent()
        Changes method from a do null event (do nothing event). Child classes process method should start with
        'if self._do_nothing:
            super().process()
        else:'
        or
        'if self._do_nothing:
            pass
        else:'

    """

    def __init__(self, name, times):
        if not isinstance(name, str):
            raise TypeError('A name given to an event should be a string.')
        self.name = name
        if isinstance(times, (float, int)):
            self.times = set([times])
        elif isinstance(times, (list, tuple, set)):
            if any(not isinstance(entry, (int, float)) for entry in times):
                raise TypeError('All entries in event_information "times" should be ints or floats.')
            self.times = set(times)
        elif isinstance(times, range):
            self.times = set(times)
        elif isinstance(times, (np.ndarray, np.generic)):
            self.times = set(times)
        else:
            raise TypeError('"times" type is not supported. ' +
                            'Accepted types are ranges (including numpy), single ints/floats, ' +
                            'or np.arrays, lists, tupples and sets of ints/floats.')
        self._do_nothing = False

    def process(self):
        """
        Method for carrying out event.

        Returns
        -------
        Nothing
        """
        pass

    def make_event_a_nullevent(self):
        self._do_nothing = True

    def undo_make_event_a_nullevent(self):
        self._do_nothing = False



class ValueFactorProportionChangeEvent(BaseEvent):
    """
    Parent event for events that involve changing a value or values.

    Parameters & Attributes
    -----------------------
    name : str
        Name given to event.
    times : floats/ints, ranges or list/tuple/set of floats/ints
        Times at which event occurs.
    value : foat/int, mutually exclusive with factor and proportion.
        Value that overrides original value(s).
    proportion : foat/int, mutually exclusive with value and factor.
        Proportion the multiplies original value(s). Must be between 0 and 1.
    factor : foat/int, mutually exclusive with value and proportion.
        Factor the multiplies original value(s).

    Methods
    -------
    process()
        Method for carrying out event.
    make_event_a_nullevent()
        Changes method into a do null event (do nothing event).
    undo_make_event_a_nullevent()
        Changes method from a do null event (do nothing event).

    """
    def __init__(self, name, times, value=None, factor=None, proportion=None):
        self._value = None
        self._factor = None
        self._proportion = None
        error_msg = 'Only one out of factor, value and proportion can be given.'
        if factor is not None and value is not None and proportion is not None:
            raise AssertionError(error_msg)
        if factor is not None and value is not None:
            raise AssertionError(error_msg)
        if factor is not None and proportion is not None:
            raise AssertionError(error_msg)
        if value is not None and proportion is not None:
            raise AssertionError(error_msg)
        if factor is not None:
            self.factor = factor
        if value is not None:
            self.value = value
        if proportion is not None:
            self.proportion = proportion
        super().__init__(name=name, times=times)
    @property
    def proportion(self):
        return self._proportion
    @proportion.setter
    def proportion(self, proportion):
        if not isinstance(proportion, Number):
            raise TypeError('Value for proportion must be a numeric type.')
        if proportion <= 0:
            raise ValueError('Value of ' + str(proportion) + ' entered for proportion must be greater than 0.' +
                             'To turn event to a nullevent (an event that transfers nothing use method "make_event_a_nullevent".')
        if proportion > 1:
            raise ValueError('Value of ' + str(proportion) +
                             ' entered for proportion must be less than or equal to 1.')
        if self.value is not None:
            warn(self.name + ' was set to value, it will now be set to a proportion.')
            self._value = None
        if self.factor is not None:
            warn(self.name + ' was set to factor, it will now be set to a proportion.')
            self._factor = None
        self._proportion = proportion
    @property
    def factor(self):
        return self._factor
    @factor.setter
    def factor(self, factor):
        if not isinstance(factor, Number):
            raise TypeError('Value for factor must be a numeric type.')
        if self.value is not None:
            warn(self.name + ' was set to value, it will now be set to a factor.')
            self._value = None
        if self.proportion is not None:
            warn(self.name + ' was set to proportion, it will now be set to a factor.')
            self._proportion = None
        self._factor = factor
    @property
    def value(self):
        return self._value
    @value.setter
    def value(self, value):
        if not isinstance(value, Number):
            raise TypeError('Value for value must be a numeric type.')
        if self.factor is not None:
            warn(self.name + ' was set to factor, it will now be set to a value.')
            self._factor = None
        if self.proportion is not None:
            warn(self.name + ' was set to proportion, it will now be set to a value.')
            self._proportion = None
        self._value = value


class TransferEvent(ValueFactorProportionChangeEvent):
    """
    Transfers values between state variables.

    Note if value, factor and proportion are all None event is a null event (do nothing event).

    Parameters & Attributes
    -----------------------
    name : str
        Name given to event.
    times : floats/ints, ranges or list/tuple/set of floats/ints
        Times at which event occurs.
    value : foat/int, optional but mutually exclusive with factor and proportion.
        Value that overrides original parameter value(s).
    proportion : foat/int, optional but mutually exclusive with value and factor.
        Proportion the multiplies original parameter value(s). Must be between 0 and 1.
    factor : foat/int, optional but mutually exclusive with value and proportion.
        Factor the multiplies original parameter value(s).
    from_index: list-like of ints
        Indexes of stat variables from which transfers take place.
    to_index: list-like of ints
        Indexes of stat variables to which transfers take place.


    Methods
    -------
    process()
        Process event by transferring amount between state variables
    make_event_a_nullevent()
        Changes method into a do null event (do nothing event).

    """
    def __init__(self, name, times, value=None, factor=None,  proportion=None,
                 from_index=None, to_index=None):
        super().__init__(name=name, times=times, value=value, factor=factor, proportion=proportion)
        if from_index is None and to_index is None:
            raise AssertionError('A container of ints must be given for from_index or to_index or both.')
        if from_index is not None:
            if not all(isinstance(index, int) for index in from_index):
                raise TypeError('All values in from_index must be ints.')
            if from_index == to_index:
                raise AssertionError('from_index and to_index must not be equivelant.')
            self.number_of_elements = len(from_index)
        if to_index is not None:
            if not all(isinstance(index, int) for index in from_index):
                raise TypeError('All values in to_index must be ints.')
            self.number_of_elements = len(from_index)
        if from_index is not None and to_index is not None and len(from_index)!=len(to_index):
            raise AssertionError('If both from_index and to_index are given they must be of the same length.')
        self.from_index = from_index
        self.to_index = to_index

    def process(self, solution_at_t, time, return_total_effected=True):
        """
        Process event by transferring amount between state variables.
        Note if value, factor and proportion are all None event is a null event (do nothing event).

        Parameters
        ----------
        solution_at_t : numpy.array
            State variable values at time t.
        time : float
            Time at t.
        return_total_effected : bool, default is True
            If true total number transferred is returned.

        Returns
        -------
        If return_total_effected is True, total number transferred is returned.
        """
        if self._do_nothing:
            super().process()
        elif self.value is None and self.factor is None and self.proportion is None:
            pass
        else:
            if self.value is not None:
                transfers = np.repeat(self.value, self.number_of_elements)
                if self.from_index is not None:
                    less_than_array = solution_at_t < transfers
                    if any(less_than_array):
                        warn('The total in one or more states was less than default value being deducted'+
                             ','+str(self.value)+', at time ' + str(time) + '.'+
                             ' Removed total population of effected state or states instead.')
                        transfers[less_than_array] = solution_at_t[self.from_index[less_than_array]]
            if self.proportion is not None:
                transfers = solution_at_t[self.from_index] * self.proportion
            if self.factor is not None:
                transfers = solution_at_t[self.from_index] * self.factor
                if self.factor > 1:
                    warn('More people are being transferred than in population.')
            if self.to_index is not None:
                solution_at_t[self.to_index] += transfers
            if self.from_index is not None:
                solution_at_t[self.from_index] -= transfers
            if return_total_effected:
                return transfers.sum()


class ChangeParametersEvent(ValueFactorProportionChangeEvent):
    """
    Event for changing parameter values.
    Note if value, factor and proportion are all None event is a null event (do nothing event).

    Parameters & Attributes
    -----------------------
    name : str
        Name given to event.
    times : floats/ints, ranges or list/tuple/set of floats/ints
        Times at which event occurs.
    changing_parameters : list-like of strings
        Parameters whose values will be changed.
    value : foat/int, optional but mutually exclusive with factor and proportion.
        Value that overrides original parameter value(s).
    proportion : foat/int, optional but mutually exclusive with value and factor.
        Proportion the multiplies original parameter value(s). Must be between 0 and 1.
    factor : foat/int, optional but mutually exclusive with value and proportion.
        Factor the multiplies original parameter value(s).

    Methods
    -------
    process()
        Process event by change parameter values for parameters listed in self.changing_parameters.
    make_event_a_nullevent()
        Changes method into a do null event (do nothing event).

    """
    def __init__(self, name, times, changing_parameters, value=None, factor=None, proportion=None):
        super().__init__(name=name, times=times, value=value, factor=factor, proportion=proportion)
        self.changing_parameters = changing_parameters

    def process(self, model_object, parameters_attribute, parameters):
        """
        Process event by change parameter values for parameters listed in self.changing_parameters.
        Note if value, factor and proportion are all None event is a null event (do nothing event).

        Parameters
        ----------
        model_object : object
            Object used to define and simulate model.
        parameters_attribute : string
            Attribute of model_object that sets parameters (must accept dictionary where keys are strings and values are
            floats/ints).
        parameters : dict {str : floats/ints}
            Parameters being used by model.


        Returns
        -------
        parameters : dict {str : floats/ints}
            Parameters being used by model after the values of those is list self.changing_parameters is changed.
        """
        if self._do_nothing:
            super().process()
        elif self.value is None and self.factor is None and self.proportion is None:
            pass
        else:
            for parameter in self.changing_parameters:
                if self.value is not None:
                    parameters[parameter] = self.value
                if self.factor is not None:
                    parameters[parameter] *= self.factor
                if self.proportion is not None:
                    parameters[parameter] *= self.proportion

            setattr(model_object, parameters_attribute, parameters)
            return parameters


class ParametersEqualSubPopEvent(BaseEvent):
    """
    Change parameters to be equal to the sum of a subpopulation.

    Parameters & Attributes
    -----------------------
    name : str
        Name given to event.
    times : floats/ints, ranges or list/tuple/set of floats/ints
        Times at which event occurs.
    changing_parameters : list-like of strings
        Parameters whose values will be changed.
    subpopulation_index : list-like of ints
        Index of subpopulation the sum of which will change self.changing_parameters to.


    Methods
    -------
    process()
        Process event by changing parameters to be equal to the sum of a subpopulation.
    make_event_a_nullevent()
        Changes method into a do null event (do nothing event).

    """
    def __init__(self, name, times, changing_parameters, subpopulation_index):
        super().__init__(name=name, times=times)
        self.changing_parameters = changing_parameters
        self.subpopulation_index = subpopulation_index

    def process(self, model_object, parameters_attribute, parameters, solution_at_t):
        """
        Process event by changing parameters to be equal to the sum of a subpopulation.

        Parameters
        ----------
        model_object : object
            Object used to define and simulate model.
        parameters_attribute : string
            Attribute of model_object that sets parameters (must accept dictionary where keys are strings and values are
            floats/ints).
        parameters : dict {str : floats/ints}
            Parameters being used by model.
        solution_at_t : numpy.array
            State variable values at time t.

        Returns
        -------
        Nothing
        """
        if self._do_nothing:
            super().process()
        else:
            value = solution_at_t[self.subpopulation_index].sum()
            for parameter in self.changing_parameters:
                parameters[parameter] = value

            setattr(model_object, parameters_attribute, parameters)
            return parameters


if __name__ == "__main__":
    pass


"""
Creation:
    Author: Martin Grunnill
    Date: 2022/09/13
Description: Classes for Multnomial random draw seeding of infections.

Classes
-------
MultnomialSeeder
    Makes multinomial draws selecting an infectious hosts branch and then state.
    
"""
from numbers import Number
import numpy as np
import math

class _InfectionBranch:
    """
    Makes multinomial draws for selecting which stage of an infection pathway to place infected hosts.
    Calculates normalised weighting of an infection branch's states, based on inverse outflow for states.

    Parameters & Attributes
    -----------------------
    name : string
        Name of branch.
    outflows: dictionary {str or ints: string}
        Keys are name or number given to state. Values are name given to parameter.

    Methods
    -------
    calculate_weighting(parameters)
        Calculate normalised weighting for each state.
    seed_infections(self, n, parameters)
        Make multinomial draw to select infectious stages of this branch to seed infection into.

    """
    def __init__(self, name, outflows):
        if not isinstance(name, str):
            raise TypeError('name argument should be a string.')
        self.name = name
        outflows_err_msg = ('outflows argument should be a dictionary,'+
                            ' with keys being strings or integers and values being string.')
        if not isinstance(outflows,dict):
            raise TypeError(outflows_err_msg)
        if any(not isinstance(key,(int,str)) for key in outflows.keys()):
            raise TypeError(outflows_err_msg)
        if any(not isinstance(value,str) for value in outflows.values()):
            raise TypeError(outflows_err_msg)
        self.outflows = outflows

    def calculate_weighting(self, parameters):
        """
        Calculate normalised weighting for each state.

        Parameters
        ----------
        parameters : dict {str: Number}
            Dictionary of parameter values.

        Returns
        -------
        noramlised_weightings : dict {str: float}
            Dictionary normalised weighting for each state.
        """
        parameters_error = ('parameters argument should be a dictionary,' +
                            ' with keys being strings and values being numbers.')
        if not isinstance(parameters, dict):
            raise TypeError(parameters_error)
        if any(not isinstance(value, Number) for value in parameters.values()):
            raise TypeError(parameters_error)
        if any(not isinstance(key,str) for key in parameters.keys()):
            raise TypeError(parameters_error)

        weightings = {}
        total = 0
        for state, outflow in self.outflows.items():
            weighting = parameters[outflow] ** -1
            weightings[state] = weighting
            total += weighting

        noramlised_weightings = {state: weight/total for state, weight in weightings.items()}

        return noramlised_weightings

    def seed_infections(self, n, parameters, rng=None):
        """
        Make multinomial draw to select infectious stages of this branch to seed infection into.

        Parameters
        ----------
        n : int
            Number of infections to seed.
        parameters : dict {str: Number}
            Dictionary of parameter values.
        rng : numpy random number generator, optional.
            Random number generator to use.

        Returns
        -------
        draw_dict : dict {str: int}
            Keys are states values are number of infections in state.
        """
        if rng is None:
            rng = np.random.default_rng()
        weighting = self.calculate_weighting(parameters)
        pvals = list(weighting.values())
        states = list(weighting.keys())
        draw = rng.multinomial(n=n, pvals=pvals, size=1)
        draw = draw[0]
        draw_dict = {state: draw[index] for index, state in enumerate(states)}
        return draw_dict


class MultnomialSeeder:
    """
    Makes multinomial draws selecting an infectious hosts branch and then state.

    Parameters
    ----------
    branch_info : nested dict
        First level keys are branches (str).
            Second level keys are states (str or ints) and values are the names of parameters controlling outflows for
            those states (str).

    Attributes
    ----------
    branches : dict {str: InfectionBranch}
        Infection branches that a host can be placed upon.
    parameters : set of strings
        Parameters (outflows) given in branch_info.

    Methods
    -------
    seed_infections(n, branch_probability, parameters)
        Draw selection of states to place infected hosts.

    """

    def __init__(self, branch_info):
        if not isinstance(branch_info,dict):
            raise TypeError('branch_info should be a dictionary.')
        self.branches = {}
        self.parameters = set()
        self.rng = None
        for branch_name, outflows in branch_info.items():
            self.parameters.update(list(outflows.values()))
            if not isinstance(branch_info, dict):
                raise TypeError('branch_info should be a dictionary of dictionaries.')
            self.branches[branch_name] = _InfectionBranch(branch_name, outflows)

    def set_seed(self,seed):
        """
        Sets random number generator seed.

        Parameters
        ----------
        seed : int (>0)

        """
        self.rng = np.random.default_rng(seed)

    def _seed_branches(self, n, branch_probability):
        """
        Make multinomial draw for which infection branch to place a host.

        Parameters
        ----------
        n : int
            Number of infections to seed.
        branch_probability : dict {string, float}
            Probability of being on each infection branch.

        Returns
        -------
        draw_dict : dict {str: int}
            Keys are branches values are number of infections on branch.
        """
        if self.rng is None:
            rng = np.random.default_rng()
        else:
            rng = self.rng
        pvals = list(branch_probability.values())
        branches = list(branch_probability.keys())
        draw = rng.multinomial(n=n, pvals=pvals, size=1)
        draw = draw[0]
        draw_dict = {branch: draw[index] for index, branch in enumerate(branches)}
        return draw_dict
    def seed_infections(self, n, branch_probability, parameters):
        """
        Draw selection of states to place infected hosts.

        Parameters
        ----------
        n : int
            Number of infections to seed.
        branch_probability : dict {string, float}
            Probability of being on each infection branch.
        parameters : dict {str: Number}
            Dictionary of parameter values.

        Returns
        -------
        infections_draw : dict {str: int}
             Keys are infected states values are number of hosts in state.
        """
        prob_error = ', all proportion argument should be a number <=1 and >=0.'
        for key, value in branch_probability.items():
            if not isinstance(value, Number):
                raise TypeError(key+' not a Number type'+ prob_error)
            if value > 1 or value < 0:
                raise ValueError(key+' is of value '+ str(value) + prob_error)
        proportions_total = sum(branch_probability.values())
        if not math.isclose(1, proportions_total, abs_tol=0.000001):
            raise ValueError('The sum of dictionary values in proportions should equal 1, it is equal to ' +
                             str(proportions_total)+'.')
        branch_draw = self._seed_branches(n, branch_probability)
        infections_draw = {}
        for branch_name, branch_seed in branch_draw.items():
            branch = self.branches[branch_name]
            branch_infection_draw = branch.seed_infections(branch_seed, parameters, self.rng)
            states_already_drawn = set(infections_draw.keys()).union(set(branch_infection_draw.keys()))
            updated_infection_draws = {state: branch_infection_draw.get(state, 0) + infections_draw .get(state, 0)
                                       for state in states_already_drawn}
            infections_draw = updated_infection_draws
        return infections_draw


if __name__ == "__main__":
    pass


"""
Creation:
    Author: Martin Grunnill
    Date: 2022-10-05
Description: Assess if a Latin-Hypercube sample size is reasonable.
"""

import pandas as pd
import os
from scipy.stats import qmc
import timeit
from tqdm import tqdm


def determine_lh_sample_size(parameters_df,
                             model_run_method,
                             start_n,
                             repeats_per_n,
                             std_aim,
                             LHS_PRCC_method,
                             save_dir,
                             attempts_to_make=float('inf'),
                             n_increase_addition = None,
                             n_increase_multiple = None,
                             y0=None,
                             other_samples_to_repeat=None,
                             max_workers=None):
    """
    Assess if a Latin-Hypercube (LH) sample size is reasonable.

    Method is as follows:
        1.  For a given LH sample size draw and simulate a model through repeats_per_n number of LHSs.
        2.  Determine standard deviation in PRCCs. If any PRCC is greater than std_aim
            increase LH sample size by either addittion (n_increase_addition)
            or multiplication (n_increase_multiple) and return to 1.

    Parameters
    ----------
    parameters_df : pd.DataFrame
        DataFrame outlining the boundaries for each parameter. Must contain fields 'Lower Bound' and
        'Upper Bound'. Name of the parameters is assumed to be in the index.
    model_run_method : function
        Method for simulating model.
    start_n : int
        Suggested sample size to start from.
    repeats_per_n : int
        Number of repeats per sample size.
    std_aim : float
        Standard deviation goal.
    LHS_PRCC_method : function
        Method for drawing LHs simulating models through LHs and determining PRCC.
    save_dir : string
        Directory for saving results.
    attempts_to_make : int
        Number of attempts to make.
    n_increase_addition : int, mutually exclusive with n_increase_multiple
        Increase in sample size, by addition, to be made if standard deviation in any paramters PRCC
        is greater than std_aim.
    n_increase_multiple : int, mutually exclusive with n_increase_addition
        Increase in sample size, by multiplication, to be made if standard deviation in any paramters PRCC
        is greater than std_aim.
    y0 : numpy.array, optional
        Intial values of state variables.
    other_samples_to_repeat : pandas.DataFrame
        Samples to resampled and merged with LH samples.
    max_workers : int, optional
        Number workers if LHS_PRCC method is parallelized.

    Returns
    -------
    int
        Sample size assessed as being reasonable.
    """
    if n_increase_addition is not None:
        if not isinstance(n_increase_addition,int):
            if isinstance(n_increase_addition, float):
                if not n_increase_addition.is_integer():
                    raise TypeError('n_increase_addition must be an interger > 0.')
        if n_increase_addition <0:
            raise ValueError('n_increase_addition must be > 0.')
    elif n_increase_multiple is not None:
        if not isinstance(n_increase_multiple, (int, float)):
            raise TypeError('n_increase_multiple must be an int or float > 1.')
        if n_increase_multiple <1:
            raise ValueError('n_increase_multiple must be > 1.')
    else:
        raise AssertionError('Either n_increase_addition or n_increase_multiple must be given.')

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    LHS_obj = qmc.LatinHypercube(len(parameters_df))
    sample_size = start_n
    aim_reached = False
    attempts_made = 0

    while attempts_made < attempts_to_make and not aim_reached:
        prcc_measures = []
        run_times = []
        save_prcc_stats_file = (save_dir +
                                '/PRCC descriptive stats at sample size ' +
                                str(sample_size) + '.csv')
        run_time_stats_file = (save_dir +
                                '/Run time descriptive stats at sample size ' +
                                str(sample_size) + '.csv')
        temporary_folder = (save_dir +
                            '/PRCC results from LH sample size ' +
                            str(sample_size))
        run_time_record_file = (save_dir +
                                '/Run times at sample size ' +
                                str(sample_size) + '.csv')
        if not os.path.exists(temporary_folder):
            os.makedirs(temporary_folder)
        if not os.path.isfile(save_prcc_stats_file):
            range_of_repeats = range(repeats_per_n)
            for repeat_num in tqdm(range_of_repeats, leave=False, position=0, colour='blue',
                                   desc='LHS resample sample size of '+str(sample_size)):
                temporary_results_file = temporary_folder + '/PRCC results from run ' + str(repeat_num)+'.csv'
                if os.path.exists(temporary_results_file):
                    prcc_measure_enty = pd.read_csv(temporary_results_file, index_col=0)
                    run_times_df = pd.read_csv(run_time_record_file, index_col=0)
                    run_times = run_times_df.Times.to_list()
                else:
                    start = timeit.default_timer()
                    if other_samples_to_repeat is not None:
                        prcc_measure_enty = LHS_PRCC_method(parameters_df, sample_size, model_run_method,
                                                            LHS_obj=LHS_obj, y0=y0,
                                                            other_samples_to_repeat=other_samples_to_repeat,
                                                            max_workers=max_workers)
                    else:
                        prcc_measure_enty = LHS_PRCC_method(parameters_df, sample_size, model_run_method,
                                                            LHS_obj=LHS_obj, y0=y0, max_workers=max_workers)
                    end = timeit.default_timer()
                    run_times.append(end - start)
                    run_times_df = pd.DataFrame({'Times': run_times})
                    run_times_df.to_csv(run_time_record_file)
                    prcc_measure_enty.to_csv(temporary_results_file)
                prcc_measures.append(prcc_measure_enty['r'])
            prcc_measures_df = pd.concat(prcc_measures, axis=1)
            prcc_measures_df.columns = range_of_repeats
            prcc_measures_df = prcc_measures_df.transpose(copy=True)
            prcc_decriptive_stats = prcc_measures_df.describe()
            prcc_decriptive_stats.to_csv(save_prcc_stats_file)
            times_decriptive_stats = run_times_df.Times.describe()
            times_decriptive_stats.to_csv(run_time_stats_file)
        else:
            prcc_decriptive_stats = pd.read_csv(save_prcc_stats_file, index_col=0)

        max_std = prcc_decriptive_stats.loc['std', :].max()
        if max_std > std_aim:
            if n_increase_addition is not None:
                sample_size += n_increase_addition
            elif n_increase_multiple is not None:
                sample_size *= n_increase_multiple

        else:
           aim_reached = True
        attempts_made += 1

    return sample_size

if __name__ == "__main__":
    pass


"""
Creation:
    Author: Martin Grunnill
    Date: 2024-02-09
Description: 
    
"""
from .events import (
    BaseEvent,
    ValueFactorProportionChangeEvent,
    TransferEvent,
    ChangeParametersEvent,
    ParametersEqualSubPopEvent,
    )
from .event_que import EventQueue
